from __future__ import annotations

from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorNoopBase import DelegatorNoopBase


class DelegatorNoopEmpty(DelegatorNoopBase):
    """
    The purpose of this delegator is NOT to have implementation of any function.
    """
