from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract


class DelegatorCheckEnvBase(DelegatorSingleFuncAbstract):
    """
    Implements base functionality used by FS_36_17_84_44 `check_env`.
    """
